 (**********************************************************************
  *                        Program REGMINT.PAS                         *
  *                            Version 1.0                             *
  *                    (c) J. Debord, January 2004                     *
  **********************************************************************
  This program performs a nonlinear least squares fit of the integrated
  Michaelis equation:

  y = H * {s0 - Km * W[s0 / Km * exp(s0 / Km - ks * e0 * t)]}

  y  = instrument response (e. g. absorbance) at time t
  H  = proportionality constant (e. g. extinction coefficient)
  s0 = initial substrate concentration
  Km = Michaelis constant
  ks = specificity constant = kcat / Km
  e0 = total enzyme concentration

  W is Lambert's function (reciprocal of x * exp(x))

  The independent variable x may be t, s0 or e0.

  H may be fitted or held constant.

  The following parameters are passed on the command line :

    1st parameter = Name of input file (default extension = .DAT)
                    The structure of the input file is described in
                    REG_IN.INC
    2nd parameter = s0   (0 if s0 varies)
    3rd parameter = e0   (0 if e0 varies)
    4th parameter = Tmax (incubation time, 0 if t varies)
    5th parameter = H
    6th parameter = 1 if H must be fitted, 0 otherwise


  There are 2 examples from the author's laboratory:

    BNA.DAT : Hydrolysis of beta-naphthylacetate by cholinesterase,
              absorbance versus time. Use: regmint bna 1 0.1 0 1.5 1

    CHE.DAT : Hydrolysis of butyrylthiocholine by cholinesterase,
              product concentration versus initial substrate conc.
              Use: regmint che 0 1 1 1 0
  **********************************************************************)

program RegMint;

uses
  Graph, FMath, Matrices, Optim, Regress, Models, PaString, PlotVar, Plot, TexPlot;

const
  MAXITER = 1000;    { Maximal number of iterations }
  TOL     = 1.0E-4;  { Required parameter precision }

var
  InFName      : String;    { Name of input file }
  Title        : String;    { Title of study }
  XName, YName : String;    { Variable names }
  N            : Integer;   { Number of points }
  X, Y         : PVector;   { Point coordinates }
  U            : PMatrix;   { Matrix of independent variables (not used here) }
  Ycalc        : PVector;   { Expected Y values }
  S            : PVector;   { Standard deviations of Y values }
  CstPar       : PVector;   { Constant parameters }
  B            : PVector;   { regression parameters }
  B_min, B_max : PVector;   { Parameter bounds }
  V            : PMatrix;   { Variance-covariance matrix of regression parameters }
  Theta        : PVector;   { Variance parameters }
  RegTest      : TRegTest;  { regression tests }
  ErrCode      : Integer;   { Error code }

  procedure ReadCmdLine(var InFName : String; var CstPar : PVector);
{ --------------------------------------------------------------------
  Reads command line parameters. Stores constant parameters in CstPar,
  such that :

    CstPar^[0] = s0   (0 if s0 varies)
    CstPar^[1] = e0   (0 if e0 varies)
    CstPar^[2] = Tmax (0 if t varies)
    CstPar^[3] = H
    CstPar^[4] = 1 if H must be fitted

  The contents of CstPar are defined in the unit FITMINT.PAS,
  in the subdirectory REG of the TP Math units directory.
  -------------------------------------------------------------------- }
  var
    I, ErrCode      : Integer;
    S0, E0, Tmax, H : Float;
  begin
    DimVector(CstPar, 4);

    { Name of input file }
    InFName := ParamStr(1);
    if Pos('.', InFName) = 0 then InFName := InFName + '.dat';

    { Initial substrate concentration }
    Val(ParamStr(2), S0, ErrCode);
    if (ErrCode <> 0) or (S0 < 0.0) then S0 := 0.0;
    CstPar^[0] := S0;

    { Enzyme concentration }
    Val(ParamStr(3), E0, ErrCode);
    if (ErrCode <> 0) or (E0 < 0.0) then E0 := 1.0;
    CstPar^[1] := E0;

    { Incubation time }
    Val(ParamStr(4), Tmax, ErrCode);
    if (ErrCode <> 0) or (Tmax < 0.0) then Tmax := 1.0;
    CstPar^[2] := Tmax;

    { Proportionality factor }
    Val(ParamStr(5), H, ErrCode);
    if (ErrCode <> 0) or (H < 0.0) then H := 1.0;
    CstPar^[3] := H;

    { Presence of constant term }
    I := 0;
    Val(ParamStr(6), I, ErrCode);
    CstPar^[4] := I;
  end;

  procedure SetParamBounds(B_min, B_max : PVector);
  { Set parameter bounds }
  var
    I : Integer;
  begin
    for I := FirstParam to LastParam do
      begin
        B_min^[I] := 0.0;
        B_max^[I] := 1.0E+10;
      end;
  end;

  {$I REG_IN.INC}    { Read input file }

  {$I REG_OUT.INC}   { Write output file }

  {$I REG_PLOT.INC}  { Plot function }

{ *************************** Main program ***************************** }

begin
  { Read command line parameters }
  ReadCmdLine(InFName, CstPar);

  { Read input file }
  if ReadInputFile(InFName, Title, XName, YName, N, X, Y) <> 0 then
    begin
      WriteLn('Error reading file ', InFName);
      Halt;
    end;

  { Initialize regression and variance models.
    See MODELS.PAS in the UNIT\REG subdirectory
    for a list of available models }
  InitModel(REG_MINT,
            VAR_CONST,  { Here we use a constant variance }
            CstPar);

  { Dimension arrays.
    Note: the variance parameters Theta^[1]..Theta^[LastVarParam]
    must be supplied if we use a non-constant variance model }
  DimVector(Theta, LastVarParam);
  DimVector(B, LastParam);
  DimVector(B_min, LastParam);
  DimVector(B_max, LastParam);
  DimMatrix(V, LastParam, LastParam);
  DimVector(Ycalc, N);
  DimVector(S, N);

  { Set parameter bounds }
  SetParamBounds(B_min, B_max);

  { Perform regression }
  ErrCode := WLSFit(X, U, Y, N, True, MAXITER, TOL, Theta,
                    B, B_min, B_max, V, Ycalc, S, RegTest);

  { Path to the graphic drivers (Default = C:\TP\BGI) }
  BGIPath := 'C:\BP\BGI';

  { Write results & plot curve }
  case ErrCode of
    OPT_OK         : begin
                       PlotGraph(Title, X, Y, N);
                       WriteOutputFile(InFName, Title, XName, YName,
                                       N, Y, Ycalc, S, B, V, RegTest);
                     end;
    OPT_SING       : WriteLn('Singular Hessian matrix !');
    OPT_BIG_LAMBDA : WriteLn('Too high Marquardt''s parameter !');
    OPT_NON_CONV   : WriteLn('Non-convergence !');
  end;
end.
