{ **********************************************************************
  *                            Unit LAMBERT.PAS                        *
  *                              Version 1.0                           *
  *                      (c) J. Debord, January 2004                   *
  **********************************************************************
  Unit for computing Lambert's W function
  This unit is translated from a Fortran program written by Barry et al.
  (http://www.netlib.org/toms/743)
  ********************************************************************** }

unit Lambert;

interface

uses
  FMath;

var
  W_Error : Integer;

function LambertW(X : Float; UBranch, Offset : Boolean) : Float;
{ ----------------------------------------------------------------------
  Lambert's W function: Y = W(X) ==> X = Y * Exp(Y)    X >= -1/e
  ----------------------------------------------------------------------
  X       = Argument
  UBranch = TRUE  for computing the upper branch (X >= -1/e, W(X) >= -1)
            FALSE for computing the lower branch (-1/e <= X < 0, W(X) <= -1)
  Offset  = TRUE  for computing W(X - 1/e), X >= 0
            FALSE for computing W(X)
  ---------------------------------------------------------------------- }

implementation

var
  NBITS : Integer;

var
  LC : record
         EM, EM9, C13, C23, EM2, D12, X0, X1,
         AN3, AN4, AN5, AN6, S21, S22, S23 : Float;
       end;

  function LambertW(X : Float; UBranch, Offset : Boolean) : Float;
  var
    I, NITER : Integer;
    AN2, DELX, ETA, RETA, T, TEMP, TEMP2, TS, WAPR, XX, ZL, ZN : Float;

  begin
    W_Error := 0;

    if Offset then
      begin
        DELX := X;
        if DELX < 0.0 then
          begin
            W_Error := 1;
            Exit;
          end;
        XX := X + LC.EM;
      end
    else
      begin
        if X < LC.EM then
          begin
            W_Error := 1;
            Exit;
          end;

        if X = LC.EM then
          begin
            LambertW := - 1.0;
            Exit;
          end;

        XX := X;
        DELX := XX - LC.EM;
      end;

    if UBranch then
      begin
        if Abs(XX) <= LC.X0 then
          begin
            LambertW := XX / (1.0 + XX / (1.0 + XX / (2.0 + XX / (0.6 + 0.34 * XX))));
            Exit;
          end;

        if XX <= LC.X1 then
          begin
            RETA := Sqrt(LC.D12 * DELX);
            LambertW := RETA / (1.0 + RETA / (3.0 + RETA / (RETA / (LC.AN4 +
                        RETA / (RETA * LC.AN6 + LC.AN5)) + LC.AN3))) - 1.0;
            Exit;
          end;

        if XX <= 20.0 then
          begin
            RETA := SQRT2 * Sqrt(1.0 - XX / LC.EM);
            AN2 := 4.612634277343749 * Sqrt(Sqrt(RETA + 1.09556884765625));
            WAPR := RETA / (1.0 + RETA / (3.0 + (LC.S21 * AN2 + LC.S22) * RETA / (LC.S23 * (AN2 + RETA)))) - 1.0;
          end
        else
          begin
            {$ifdef CPU387}
            ZL := fLn(XX);
            WAPR := fLn(XX / fLn(Power(XX / ZL, fExp(- 1.124491989777808 / (0.4225028202459761 + ZL)))));
            {$else}
            ZL := Ln(XX);
            WAPR := Ln(XX / Ln(Power(XX / ZL, Exp(- 1.124491989777808 / (0.4225028202459761 + ZL)))));
            {$endif}
          end
      end
    else
      begin
        if XX >= 0.0 then
          begin
            W_Error := 1;
            Exit;
          end;

        if XX <= LC.X1 then
          begin
            RETA := Sqrt(LC.D12 * DELX);
            LambertW := RETA / (RETA / (3.0 + RETA / (RETA / (LC.AN4 +
                        RETA / (RETA * LC.AN6 - LC.AN5)) - LC.AN3)) - 1.0) - 1.0;
            Exit;
          end;

        {$ifdef CPU387}
        ZL := fLn(- XX);
        {$else}
        ZL := Ln(- XX);
        {$endif}

        if XX <= LC.EM9 then
          begin
            T := - 1.0 - ZL;
            TS := Sqrt(T);
            WAPR := ZL - (2.0 * TS) / (SQRT2 + (LC.C13 - T / (2.7E2 + TS * 127.0471381349219)) * TS);
          end
        else
          begin
            ETA := 2.0 - LC.EM2 * XX;
            {$ifdef CPU387}
            WAPR := fLn(XX / fLn(- XX / ((1.0 - 0.5043921323068457 * (ZL + 1.0)) * (Sqrt(ETA) + ETA / 3.0) + 1.0)));
            {$else}
            WAPR := Ln(XX / Ln(- XX / ((1.0 - 0.5043921323068457 * (ZL + 1.0)) * (Sqrt(ETA) + ETA / 3.0) + 1.0)));
            {$endif}
          end
      end;

    if NBITS < 56 then NITER := 1 else NITER := 2;

    for I := 1 to NITER do
      begin
        {$ifdef CPU387}
        ZN := fLn(XX / WAPR) - WAPR;
        {$else}
        ZN := Ln(XX / WAPR) - WAPR;
        {$endif}
        TEMP := 1.0 + WAPR;
        TEMP2 := TEMP + LC.C23 * ZN;
        TEMP2 := 2.0 * TEMP * TEMP2;
        WAPR := WAPR * (1.0 + (ZN / TEMP) * (TEMP2 - ZN) / (TEMP2 - 2.0 * ZN));
      end;

    LambertW := WAPR;
  end;

begin
  {$ifdef CPU387}
  NBITS := Round(- fLog2(MACHEP));
  LC.EM := - fExp(- 1.0);
  LC.EM9 := - fExp(- 9.0);
  {$else}
  NBITS := Round(- Log2(MACHEP));
  LC.EM := - Exp(- 1.0);
  LC.EM9 := - Exp(- 9.0);
  {$endif}
  with LC do
    begin
      C13 := 1.0 / 3.0;
      C23 := 2.0 * C13;
      EM2 := 2.0 / EM;
      D12 := - EM2;
      X0 := Power(MACHEP, 1.0 / 6.0) * 0.5;
      X1 := (1.0 - 17.0 * Power(MACHEP, 2.0 / 7.0)) * EM;
      AN3 := 8.0 / 3.0;
      AN4 := 135.0 / 83.0;
      AN5 := 166.0 / 39.0;
      AN6 := 3167.0 / 3549.0;
      S21 := 2.0 * SQRT2 - 3.0;
      S22 := 4.0 - 3.0 * SQRT2;
      S23 := SQRT2 - 2.0;
    end;
end.
